from multiprocessing import Pool
import itertools
import pandas as pd
import numpy as np
import networkx as nx
import datetime
import csv
import os


def clean_email_data(dir, files='all', include_subject=False, engine='c', encoding='latin', delete_old_file=False):
    """Cleans email data from Splunk to key fields only

    Parameters
    ----------
    dir: String ()
        Path to the root directory containing the data files to process

    file: List or 'all' (optional)
        List containing all files to be processed within the directory. If 'all'
        passed, then all CSV files in the directory will be processed

    include_subject: Boolean default = False
        Defines whether the Subject field should be included in the cleaned data

    engine: String default='c'
        The Pandas engine to read in the data, either 'c' or 'python'

    encoding: String default = 'latin'
        The file encoding of files to be read

    delete_old_file: Boolean default = False
        If set to True, the original Splunk data file will be delete_old_file

    Returns
    -------
    Nothing is returned by the function, but new files are written to 'dir' that
    have been cleaned
    """
    start = datetime.datetime.now()
    print(
        'Starting: ', str(start.hour).zfill(2) + ':'+ str(start.minute).zfill(2)
    )

    if files == 'all':
        files_to_process =  [f for f in os.listdir(dir) if f.endswith('.csv')]
    else:
        files_to_process = files

    num_files = len(files_to_process)
    print(f'{num_files} to process')

    cols_to_load = [
        'message_id',
        'SenderAddress',
        'RecipientAddress',
        'Received',
        'Size',
        'Status'
    ]

    if include_subject:
        cols_to_load.append('Subject')

    for idx, file in enumerate(files_to_process):
        path = dir + file

        df_emails = pd.read_csv(
            path,
            usecols=cols_to_load,
            engine=engine,
            encoding=encoding,
            error_bad_lines=False)

        output_fname = dir + file[:-4] + '_cleaned.csv'

        df_emails.to_csv(output_fname, index=False)

        if delete_old_file:
            os.unlink(path)

        finish = datetime.datetime.now()
        print(
            f'File {idx+1}/{num_files}: {file} finished at: '
            '{str(finish.hour).zfill(2)}:{str(finish.minute).zfill(2)}'
        )

    return None

def generate_node_edge_lists(email_data, demographic_data, demographic_key, output_dir, include_subject=False):
    """Generates Node and Edge lists from email data and saves them to csv

    Parameters
    ----------
    email_data: List of Strings
        List of paths to all files containing email data to be processed

    demographic_data: String
        Path to file containing all node demographic data to be added to Node
        list

    demographic_key: String
        Column in demographic_data that contains email address to act as join
        to email_data

    output_dir: String
        Path to directory to save Node and Edge lists into. Must include '/'
        at end.

    include_subject: Boolean default = False
        Defines whether the Subject field should be included in the email data

    Returns
    -------
    nodeList_fPath: String
        Path to the Node List generated

    edgeList_fPath: String
        Path to the Edge List generated
    """
    start = datetime.datetime.now()
    print('Starting: ', str(start.hour).zfill(2) + ':'+ str(start.minute).zfill(2))

    cols_to_load = [
        'SenderAddress',
        'RecipientAddress',
        'Received',
        'Size'
    ]

    if include_subject:
        cols_to_load.append('Subject')

    df_emails = pd.DataFrame(columns=cols_to_load)

    #Loading in email meta-data
    now = datetime.datetime.now()
    print(
        'Loading in email data: ',
        str(now.hour).zfill(2) + ':'+ str(now.minute).zfill(2)
    )

    for file in email_data:
        df_temp = pd.read_csv(file, usecols=cols_to_load)
        df_emails = df_emails.append(df_temp, sort=True)

    now = datetime.datetime.now()
    print('Email data loaded: ', str(now.hour).zfill(2) + ':'+ str(now.minute).zfill(2))

    #Loading in demographic data
    df_demographics = pd.read_csv(demographic_data, encoding='Latin')
    now = datetime.datetime.now()
    print('Demographic data loaded: ', str(now.hour).zfill(2) + ':'+ str(now.minute).zfill(2))

    #Cleaning up data-sets
    df_emails['RecipientAddress'] = df_emails['RecipientAddress'].str.lower()
    df_emails['SenderAddress'] = df_emails['SenderAddress'].str.lower()
    df_demographics[demographic_key] = df_demographics[demographic_key].str.lower()
    df_demographics = df_demographics.drop_duplicates(demographic_key).set_index(demographic_key)

    #Prepare Node List
    now = datetime.datetime.now()
    print('Node list preparation started: ', str(now.hour).zfill(2) + ':'+ str(now.minute).zfill(2))

    ## Generate list of unique emails
    unique_emails = list(df_emails['RecipientAddress'].unique())
    for idx, email in enumerate(df_emails['SenderAddress'].unique()):
        if email not in unique_emails:
            unique_emails.append(email)

    node_list = []
    for idx, email in enumerate(unique_emails):
        node = {
            'email': email
        }
        node_list.append(node)

    df_nodes = pd.DataFrame(node_list)

    for attr in list(df_demographics.columns):
        df_nodes[attr] = df_nodes['email'].map(df_demographics[attr])

    #Prepare Edge List
    now = datetime.datetime.now()
    print('Edge list preparation started: ', str(now.hour).zfill(2) + ':'+ str(now.minute).zfill(2))
    df_edges = pd.DataFrame({'weight': df_emails.groupby(['SenderAddress','RecipientAddress']).size()}).reset_index()

    #Save Lists to csv
    now = datetime.datetime.now()
    print('Saving Node and Edge lists: ', str(now.hour).zfill(2) + ':'+ str(now.minute).zfill(2))
    fName_date = str(now.year) + str(now.month).zfill(2) + str(now.day).zfill(2) + "_" + str(now.hour).zfill(2) + str(now.minute).zfill(2)

    nodeList_fName = 'nodeList_'+ fName_date + '.csv'
    nodeList_fPath = output_dir + nodeList_fName

    edgelist_fName = 'edgeList_'+ fName_date + '.csv'
    edgeList_fPath = output_dir + edgelist_fName

    df_nodes.to_csv(nodeList_fPath, index=False)
    df_edges.to_csv(edgeList_fPath, index=False)

    now = datetime.datetime.now()
    print('Node and Edge list preparation complete: ', str(now.hour).zfill(2) + ':'+ str(now.minute).zfill(2))
    print('Node List: ', nodeList_fName)
    print('Edge List: ', edgelist_fName)

    return nodeList_fPath, edgeList_fPath


def generate_nx_digraph(node_list, edge_list):
    """Generates NetworkX DiGraph object

    Parameters
    ----------
    node_list : String or Pandas Dataframe
        Path to file containing Node list, or Dataframe of Node List

    edge_list : String  or Pandas Dataframe
        Path to file containing Edge list

    Returns
    -------
    G: NetworkX DiGraph
        NetworkX DiGraph object
    """
    # Create our graph - directed = DiGraph
    G = nx.DiGraph()

    # Read in Node List
    print('Reading in Node list')
    with open(node_list, 'r') as nodecsv:
        nodereader = csv.reader(nodecsv)
        node_list = [n for n in nodereader]
        node_attrs = node_list[:1][0]
        nodes = node_list[1:]

    node_names = [n[0] for n in nodes]
    G.add_nodes_from(node_names)

    for idx, attr in enumerate(node_attrs[1:], 1):
        temp_dict = {}
        for node in nodes:
            temp_dict[node[0]] = node[idx]

        nx.set_node_attributes(G, temp_dict, attr)

    # Read in the edgelist file
    print('Node list loaded; Reading in Edge list')
    with open(edge_list, 'r') as edgecsv:
        edgereader = csv.reader(edgecsv)
        edges = [tuple(e[0:]) for e in edgereader][1:]

    G.add_weighted_edges_from(edges)

    print('NetworkX Graph generated')

    return G

def generate_nx_digraph_pandas(nodes_df, edges_df):
    node_file = './nodes_temp.csv'
    edge_file = './edge_temp.csv'

    nodes_df.to_csv(node_file, index=False)
    edges_df.to_csv(edge_file, index=False)

    G = generate_nx_digraph(node_file, edge_file)

    os.remove(node_file)
    os.remove(edge_file)

    return G


def _get_density(df_nodes, row, target_attribute):
    if row['sender_group'] == row['recipient_group']:
        node_count = df_nodes[df_nodes[target_attribute]==row['recipient_group']]['email'].count()
        density = row['num_edges']/(node_count * (node_count - 1))
        return density
    else:
        node_count_sender = df_nodes[df_nodes[target_attribute]==row['sender_group']]['email'].count()
        node_count_recipient = df_nodes[df_nodes[target_attribute]==row['recipient_group']]['email'].count()
        density = row['num_edges']/(node_count_sender * node_count_recipient)
        return density


def calc_density(df_nodes, df_edges, target_attribute):
    """Calculates the density of connections between the groups in a specific target target_attribute

    Parameters
    ----------
    df_nodes : Pandas DataFrame
        Dataframe containing the Node list

    df_edges : Pandas DataFrame
        DataFrame containing the edge list

    target_attribute : String
        Name of attribute in Node List that we want to calculate the densities between

    Returns
    ----------
    Pandas DataFrame containing the densities, grouped by the target_attribute values
    """

    df_edges = df_edges.copy()
    df_nodes = df_nodes.copy()

    df_edges['sender_group'] = df_edges['SenderAddress'].map(df_nodes.set_index('email')[target_attribute])
    df_edges['recipient_group'] = df_edges['RecipientAddress'].map(df_nodes.set_index('email')[target_attribute])

    df_densities = pd.DataFrame({'num_edges': df_edges.groupby(['sender_group','recipient_group'])['weight'].agg('count')}).reset_index()

    df_densities['density'] = df_densities.apply(lambda row: _get_density(df_nodes, row, target_attribute), axis=1)

    df_densities = df_densities[['sender_group','recipient_group','density']]
    df_densities['density'] = round(df_densities['density'],3)

    return df_densities


def _get_modularity(df_edges, row, target_attribute, weighted, source):
    if source == 'sender_group':
        source_group = 'sender_group'
        end_group = 'recipient_group'
    else:
        source_group = 'recipient_group'
        end_group = 'sender_group'

    if weighted:
        num_same_group = df_edges[(df_edges[source_group]==row[source_group]) & (df_edges[end_group]==row[source_group])]['weight'].count()
        num_diff_group = df_edges[(df_edges[source_group]==row[source_group]) & (df_edges[end_group]!=row[source_group])]['weight'].count()
    else:
        num_same_group = df_edges[(df_edges[source_group]==row[source_group]) & (df_edges[end_group]==row[source_group])]['weight'].sum()
        num_diff_group = df_edges[(df_edges[source_group]==row[source_group]) & (df_edges[end_group]!=row[source_group])]['weight'].sum()

    if num_same_group != 0:
        modularity = num_same_group/num_diff_group
        return modularity
    else:
        return np.nan


def calc_modularity(df_nodes, df_edges, target_attribute, weighted=False, direction='outbound'):
    """Calculates the Modularity of connections originating from groups in a specific target target_attribute

    Parameters
    ----------
    df_nodes : Pandas DataFrame
        Dataframe containing the Node list

    df_edges : Pandas DataFrame
        DataFrame containing the edge list

    target_attribute : String
        Name of attribute in Node List that we want to calculate the
        modularities between

    weighted : Boolean default False
        If set to True, the modularities will be weighted by the amount of
        email traffic.
        If False, will just calculate on basis on presence of a connection

    direction : String default = 'outbound'
        'outbound' or 'inbound' determines the direction of the email traffic
        to be considered

    Returns
    ----------
    Pandas DataFrame containing the modularities, grouped by the
    target_attribute values
    """
    if direction == 'outbound':
        source_group = 'sender_group'
        end_group = 'recipient_group'
    else:
        source_group = 'recipient_group'
        end_group = 'sender_group'

    df_edges = df_edges.copy()
    df_nodes = df_nodes.copy()

    df_edges['sender_group'] = df_edges['SenderAddress'].map(df_nodes.set_index('email')[target_attribute])
    df_edges['recipient_group'] = df_edges['RecipientAddress'].map(df_nodes.set_index('email')[target_attribute])

    if weighted:
        df_modularities = pd.DataFrame({'num_edges': df_edges.groupby([source_group])['weight'].agg('sum')}).reset_index()
    else:
        df_modularities = pd.DataFrame({'num_edges': df_edges.groupby([source_group])['weight'].agg('count')}).reset_index()

    df_modularities['modularity'] = df_modularities.apply(
        lambda row: _get_modularity(
            df_edges,
            row,
            target_attribute,
            weighted,
            source_group)
        , axis=1)

    df_modularities = df_modularities[[source_group,'modularity']]

    return df_modularities

def chunks(l, n):
    """Divide a list of nodes `l` in `n` chunks"""
    l_c = iter(l)
    while 1:
        x = tuple(itertools.islice(l_c, n))
        if not x:
            return
        yield x

def _betmap(G_normalized_weight_sources_tuple):
    """Pool for multiprocess only accepts functions with one argument.
    This function uses a tuple as its only argument. We use a named tuple for
    python 3 compatibility, and then unpack it when we send it to
    `betweenness_centrality_source`
    """
    return nx.betweenness_centrality_source(*G_normalized_weight_sources_tuple)

def betweenness_centrality_parallel(G, processes=None):
    """Parallel betweenness centrality function"""
    p = Pool(processes=processes)
    node_divisor = len(p._pool) * 4
    node_chunks = list(chunks(G.nodes(), int(G.order() / node_divisor)))
    num_chunks = len(node_chunks)
    bt_sc = p.map(_betmap,
                  zip([G] * num_chunks,
                      [True] * num_chunks,
                      [None] * num_chunks,
                      node_chunks))

    # Reduce the partial solutions
    bt_c = bt_sc[0]
    for bt in bt_sc[1:]:
        for n in bt:
            bt_c[n] += bt[n]
    return bt_c



import pandas as pd
import pkg_resources
from subprocess import call
from datetime import datetime

def csv_to_utf8(file):
    """Converts a csv file to utf-8 encoding

    Parameters
    ----------
    file: String
        Path to file to be encoded 

    Returns
    -------
    fName:
        String of path to reformatted file
    """
    df = pd.read_csv(
        file,
        encoding = 'latin'
    )

    fName = file[:-4] + '-reformatted.csv'

    df.to_csv(
        fName,
        index=False
    )

    print(f'{fName} created')

    return fName

def upgradeAllPackages():
    packages = [dist.project_name for dist in pkg_resources.working_set]

    packages_to_ignore = [
        'pycurl',
        'pywin32'
    ]

    for p in packages_to_ignore:
        if (p in packages):
            packages.remove(p)
            print(f'Skipping {p}')

    call("pip install --trusted-host pypi.org --trusted-host files.pythonhosted.org --upgrade " + ' '.join(packages), shell=True)

def installPackage(package):
    try:
        call(f"pip install --trusted-host pypi.org --trusted-host files.pythonhosted.org {package}", shell=True)
        print(f'{package} successfully installed')
    except Exception as e:
        print("Error - {}".format(e))

def getDateTimeStamp(time=True):
    """Function to generate a Date/Time stamp for the current time

    Parameters
    ----------
    time: Boolean
        Set to True to include time in Date stamp. False returns just date. Defaults to True.

    Returns
    -------
    stamp:
        String of date-time stamp
    """
    now = datetime.now()
    stamp = str(now.year) + str(now.month).zfill(2) + str(now.day).zfill(2)
    if time:
        stamp = stamp + "_" + str(now.hour).zfill(2) + str(now.minute).zfill(2)
    return stamp

def continuous_logging(msg):
    """Function to print out messages that over-write each other as a new message is sent

    Parameters
    ----------
    msg: String
        the message to be printed to the console

    Returns
    -------
    None
    """

    print(f'\r{msg}\r', end='', flush=True)

    return None



import pandas as pd
import requests
from bs4 import BeautifulSoup
from datetime import datetime
import base64
import os
import urllib3
urllib3.disable_warnings()
from fake_useragent import UserAgent

class Visier:
    """Class to manage ingestion of Visier data connectors

    See Visier user documentation for more information (you will need to log in to the service portal first): 
    https://my.visier.com/csm?id=kb_article_view&sysparm_article=KB0014851

    Parameters
    ----------
    company : str
        The company name used to connect to the Visier application e.g. [company].visier.com
    
    api_key : str
        The API key generated in the admin panel of the Visier application

    user : str
        The username used to connect to the data_connectors (typically email). This user must 
        have access to all data connectors that you intend to connect to with this Class

    pword : str
        The associated password for :code:`user`
    """
    def __init__(self, company, api_key, user, pword):

        self.company = company
        self.api_key = api_key

        self.auth = "Basic {}==".format(base64.b64encode(bytes("{}:{}".format(user,pword),"utf-8")).decode("utf-8"))

        self.url_base = f"https://{company}.api.visier.io/api/dataconnector/getData"

    def get_connector(self, connector_id, fName=None):
        """Function to make API call to Visier Data Connectors

        Parameters
        ----------
        connector_id: String
            String of id of data connector, available from Visier

        fName: String, optional
            Full file path, including csv extension that you want to give your file if you want to save it. If not declared, no file will be saved.

        Returns
        -------
        df: Pandas Dataframe
            Pandas Dataframe of returned data from the connector

        """
        urllib3.disable_warnings()
        params = {
                'id': connector_id,
                'apikey': self.api_key,
                'a':'b'
            }

        header = {"Authorization":self.auth}

        save=True
        if fName is None:
            save=False
            fName='temp.csv'

        r = requests.get(self.url_base, params=params, headers=header, verify=False)

        if r.status_code == 200:
            f=open(fName,'wb')
            f.write(r.content)
            f.close()
        else:
            raise ValueError(f'API call failed with status code: {r.status_code}')

        df = pd.read_csv(fName)

        if not save:
            os.remove(fName)
        
        return df


        

def get_random_useragent(browser=None):
    '''Function to generate a random user agent for web scraping

    Parameters
    ----------
    browser: String
        Valid values are: 'ie', 'opera', 'chrome', 'firefox', 'safari'

    Returns
    ----------
    String of valid User Agent

    '''

    ua = UserAgent()

    useragent = ua.random

    if browser is not None:
        useragent = ua[browser]

    return useragent

def get_glassdoor_ratings(glassdoor_id):
    base_url = f'https://www.glassdoor.co.uk/Reviews/{glassdoor_id}'
    reviews = []
    today = datetime.today().strftime('%Y-%m-%d')
    headers = {
        'user-agent': get_random_useragent()
    }
    urllib3.disable_warnings()

    #Overall Rating
    temp_url = base_url + '.htm'
    response = requests.get(temp_url, headers=headers, verify=False)
    soup = BeautifulSoup(response.text, 'html.parser')
    #Global Values
    country = 'Experian Global'
    rating = soup.find(
        'div',
        attrs={'class': 'common__EIReviewsRatingsStyles__ratingNum'}
    )
    num_reviews = soup.find(
        'div',
        attrs={'class': 'common__EIReviewSortBarStyles__sortsHeader'}
    )
    #Temporary Dictionary of values
    temp_dict = {
        'date': today,
        'country': country,
        'num_reviews': int(num_reviews.h2.span.strong.text),
        'rating': float(rating.text)
    }
    reviews.append(temp_dict)

    #Loop over all countries until no more found
    i=1
    keep_looping = True
    while keep_looping:
        try:
            temp_url = base_url + '-EI_IE42406.0,8_IL.9,23_IN' + str(i) + '.htm'
            response = requests.get(temp_url, headers=headers, verify=False)
            soup = BeautifulSoup(response.text, 'html.parser')

            country = soup.find(
                'div',
                attrs={'class': 'eiReviews__EIReviewsPageContainerStyles__EIReviewsPageContainer'}
            ).h1.text[9:-8]
            rating = soup.find(
                'div',
                attrs={'class': 'common__EIReviewsRatingsStyles__ratingNum'}
            )
            num_reviews = soup.find(
                'div',
                attrs={'class': 'common__EIReviewSortBarStyles__sortsHeader'}
            )

            if len(country) > 0:
                if num_reviews is not None:
                    temp_dict = {
                        'date': today,
                        'country': country,
                        'num_reviews': int(num_reviews.h2.span.strong.text),
                        'rating': float(rating.text)
                    }

                    reviews.append(temp_dict)
            else:
                keep_looping = False

            i+=1
            continue
        except Exception as e:
            keep_looping = False
            i+=1
            continue

    return pd.DataFrame(reviews)





import pandas as pd

def concat_fields(df, fields):
    """Creates the Concatenation fields for ARM

    Parameters
    ----------
    df: Pandas DataFrame
        DataFrame containing the source Data

    fields: List of Tuples
        List of Tuples containing the pairs of fields to be combined together.
        Each tuple should only length 2

    Returns
    -------
    df:
        Dataframe with new fields added
    """
    print(f'Creating {len(fields)} new fields')

    for idx, field in enumerate(fields):
        if (type(field) is tuple) and len(field)==2:
            new_field = field[0] + '_' + field[1]
            df[new_field] = df[field[0]] + '_' + df[field[1]]
            print(f'{new_field} created')
        else:
            print(
                f'Field at index {idx} skipped. It is either not a tuple or '
                'not of the correct length'
            )
    return df



def gen_expn_colors():
    """Generates the Experian brand colors

    Parameters
    ----------
    None

    Returns
    ----------
    list of Experian brand colors

    """

    expn_colours=[
        "#26478d", # Dark Blue
        "#632678", # Purple
        "#406eb3", # Light Blue
        "#982881", # Maroon
        "#ba2f7d", # Pink
        "#575756", # Grey
        # Light shaded versions
        '#7d91bb',
        '#a17dae',
        '#8ca8d1',
        '#c17eb3',
        '#d682b1',
        '#9a9a9a',
        ]

    return expn_colours

def gen_gradient_cmap(start, end, steps=50):
    """Generates a seaborn graduated color map

    Parameters
    ----------
    start: tuple
        rgb color
    end: tuple
        rgb color
    steps: int, optional
        The number of increments to return in the color gradient, default is 50

    Returns
    ----------
    list of colors for use in saeborn color maps

    """

    cmap = []
    red_step = (end[0] - start[0])/steps
    green_step = (end[1] - start[1])/steps
    blue_step = (end[2] - start[2])/steps

    start_hex = '#' + ''.join('{:02x}'.format(c) for c in start)

    cmap.append(start_hex)

    for i in range(1,steps-1):
        temp_rgb=[0,0,0]

        temp_rgb[0] = int(start[0] + i*red_step)
        temp_rgb[1] = int(start[1] + i*green_step)
        temp_rgb[2] = int(start[2] + i*blue_step)

        temp_hex = '#' + ''.join('{:02x}'.format(c) for c in temp_rgb)

        cmap.append(temp_hex)

    end_hex = '#' + ''.join('{:02x}'.format(c) for c in end)
    cmap.append(end_hex)

    return cmap



import os 
import smtplib
from datetime import datetime as datetime
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

mimetypes = {
    #images
    'png': {'type': 'image', 'extension': 'png'},
    'jpg': {'type': 'image', 'extension': 'jpg'},
    #plain files
    'csv': {'type': 'application', 'extension': 'csv'},
    'txt': {'type': 'application', 'extension': 'txt'},
    #MS Office
    'xlsx': {'type': 'application', 'extension': 'vnd.ms-excel'},
    'xls': {'type': 'application', 'extension': 'vnd.ms-excel'},
    #Other
    'pdf': {'type': 'application', 'extension': 'octate-stream'},
}

class Email:
    """Helper class to manage creating email body & headers

    This class does not currently include any authenticaion onto the email server

    Parameters
    ----------
    sender : str
        The email address you wish to appear in the 'From' field on the email.
        Can be just plain email: 'example@gmail.com', or alternatively in the format:
        'Example <example@gmail.com>'
    
    server_addr : str
        The url of the email server you are connecting to, e.g. relay.abc.local

    server_port : int
        The port through which the connection to the email server is made
    """

    def __init__(self, sender, server_addr, server_port):
        self.sender = sender
        self.server_addr = server_addr
        self.server_port = server_port

        self.server = smtplib.SMTP(
            self.server_addr, 
            self.server_port
            )

    def set_msg_templates(self, html_template, plain_template):
        self.html_template = html_template
        self.plain_template = plain_template

    def _generate_body(self, variables = {}):
        html = self.html_template
        plain = self.plain_template

        if len(variables)>0:
            html = html.format(**variables)
            plain = plain.format(**variables)

        return html, plain

    def _process_attachment(self, fpath, attachment_id):
        fname = os.path.basename(fpath)
        with open(fpath, 'rb') as f:
            ext = fname.split('.')[-1]

            mime = MIMEBase(
                mimetypes[ext]['type'],
                mimetypes[ext]['extension'],
                filename = fname)

            #Add Headers
            mime.add_header('Content-Disposition', 'attachment', filename=fname)
            mime.add_header('X-Attachment-Id', f'{attachment_id}')
            mime.add_header('Content-ID', f'<{attachment_id}>')

            mime.set_payload(f.read())
            encoders.encode_base64(mime)

            return mime 

    def send_email(self, recips, subject, cc=None, variables={}, attachments=None):
        if not isinstance(recips, list):
            recips = [recips] 

        #Set up message basics
        msg = MIMEMultipart('alternative')
        msg['From'] = self.sender
        msg['To'] = ", ".join(recips)
        msg['Subject'] = subject

        #Add recipients to CC if needed
        if cc is not None:
            if not isinstance(cc, list):
                cc = [cc]

            msg['Cc'] = ", ".join(cc)
            recips = recips + cc
        
        # Msg Body
        html,plain = self._generate_body(variables=variables)
        msg.attach(MIMEText(html, 'html'))
        msg.attach(MIMEText(plain, 'plain'))

        #TODO add default attachment handling
        for idx, attachment in enumerate(attachments):
            mime = self._process_attachment(attachment, idx+1)
            msg.attach(mime)

        self.server.sendmail(self.sender, recips, msg.as_string())
        return None
